import type { PluginFunction, PluginValidateFn } from '@graphql-codegen/plugin-helpers';
import type { RawClientSideBasePluginConfig } from '@graphql-codegen/visitor-plugin-common';
/**
 * @description This plugin is based on the TypeScript GraphQL Files Modules,
 * but generates typed document nodes, using the types generated
 * by TypedDocumentNode plugin.
 *
 * You need to run graphql-code-generator with the TypedDocumentNode plugin first
 * and set `typedDocumentNodeModule` option to the module path of the generated file.
 *
 * This plugin generates TypeScript typings for `.graphql` files containing GraphQL documents, which later on can be consumed using [`graphql-tag/loader`](https://github.com/apollographql/graphql-tag#webpack-preprocessing-with-graphql-tagloader) or use `string` types if you will use the operations as raw strings, and get type-check and type-safety for your imports. This means that any time you import objects from `.graphql` files, your IDE will provide auto-complete.
 *
 * This plugin also handles `.graphql` files containing multiple GraphQL documents, and name the imports according to the operation name.
 *
 * > ⚠ Fragments are not generated with named imports, only as default imports, due to `graphql-tag/loader` behavior.
 */
export interface TypeScriptGraphQLModuleDeclarationsPluginConfig extends RawClientSideBasePluginConfig {
    /**
     * @default ""
     * @description Allows specifying a module definition path prefix to provide distinction
     * between generated types.
     *
     * @exampleMarkdown
     * ```yml
     * generates: src/api/user-service/queries.d.ts
     *  documents: src/api/user-service/queries.graphql
     *  plugins:
     *    - typescript-graphql-files-modules
     *  config:
     *    # resulting module definition path glob: "*\/api/user-service/queries.graphql"
     *    modulePathPrefix: "/api/user-service/"
     * ```
     */
    modulePathPrefix?: string;
    /**
     * @description By default, only the filename is being used to generate TS module declarations. Setting this will generate it with a full path based on the path provided.
     */
    relativeTo: string;
    /**
     * @description Where should the file which was generated by the typed-document-node plugin be imported from.
     */
    typedDocumentNodeModule: string;
    /**
     * @default *\/
     * @description By default, a wildcard is being added as prefix, you can change that to a custom prefix
     */
    prefix?: string;
}
export declare const plugin: PluginFunction;
export declare const validate: PluginValidateFn<unknown>;
